// RangeImage.cpp: RangeImage NX̃Cve[V
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "RangeImageViewer.h"
#include "RangeImage.h"
#include "math.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

#define PI 3.14159265

//////////////////////////////////////////////////////////////////////
// \z/
//////////////////////////////////////////////////////////////////////

RangeImage::RangeImage()
{
	I = NULL;
	Ix = NULL;
	Iy = NULL;
	normal = NULL;
	normal_f = NULL;
	isRidge = NULL;
	isRavine = NULL;
	isEdge = NULL;
}

RangeImage::~RangeImage()
{
	for(int i=-1; i<sizeY+1; i++)
		I[i]--;
	I--;
	for(i=0; i<sizeY+2; i++)
		delete[] I[i];
	delete[] I;

	for(i=-1; i<sizeY+1; i++)
		Ix[i]--;
	Ix--;
	for(i=0; i<sizeY+2; i++)
		delete[] Ix[i];
	delete[] Ix;

	for(i=-1; i<sizeY+1; i++)
		Iy[i]--;
	Iy--;
	for(i=0; i<sizeY+2; i++)
		delete[] Iy[i];
	delete[] Iy;

	for(i=0; i<sizeY; i++)
		delete[] Ixx[i];
	delete[] Ixx;

	for(i=0; i<sizeY; i++)
		delete[] Ixy[i];
	delete[] Ixy;

	for(i=0; i<sizeY; i++)
		delete[] Iyx[i];
	delete[] Iyx;

	for(i=0; i<sizeY; i++)
		delete[] Iyy[i];
	delete[] Iyy;
	
	for(i=0; i<sizeY; i++)
		delete[] normal[i];
	delete[] normal;

	for(i=0; i<sizeY-1; i++)
		delete[] normal_f[i];
	delete[] normal_f;

	for(i=0; i<sizeY; i++){
		delete[] isRidge[i];
		delete[] isRavine[i];
	}
	delete[] isRidge;
	delete[] isRavine;
	delete[] isEdge;
}

void RangeImage::setSize(int sizeX, int sizeY)
{
	this->sizeX = sizeX;
	this->sizeY = sizeY;

	I = new float*[sizeY+2];
	for(int i=0; i<sizeY+2; i++)
		I[i] = new float[sizeX+2];
	I++;
	for(i=-1; i<sizeY+1; i++)
		I[i]++;

	Ix = new float*[sizeY+2];
	for(i=0; i<sizeY+2; i++)
		Ix[i] = new float[sizeX+2];
	Ix++;
	for(i=-1; i<sizeY+1; i++)
		Ix[i]++;

	Iy = new float*[sizeY+2];
	for(i=0; i<sizeY+2; i++)
		Iy[i] = new float[sizeX+2];
	Iy++;
	for(i=-1; i<sizeY+1; i++)
		Iy[i]++;

	Ixx = new float*[sizeY];
	for(i=0; i<sizeY; i++)
		Ixx[i] = new float[sizeX];

	Ixy = new float*[sizeY];
	for(i=0; i<sizeY; i++)
		Ixy[i] = new float[sizeX];

	Iyx = new float*[sizeY];
	for(i=0; i<sizeY; i++)
		Iyx[i] = new float[sizeX];

	Iyy = new float*[sizeY];
	for(i=0; i<sizeY; i++)
		Iyy[i] = new float[sizeX];

	normal = (float (**)[3])new float*[sizeY];
	for(i=0; i<sizeY; i++)
		normal[i] = new float[sizeX][3];

	normal_f = (float (**)[3])new float*[sizeY-1];
	for(i=0; i<sizeY-1; i++)
		normal_f[i] = new float[sizeX-1][3];

	isRidge = new bool*[sizeY];
	isRavine = new bool*[sizeY];
	isEdge = new bool*[sizeY];
	for(i=0; i<sizeY; i++){
		isRidge[i] = new bool[sizeX];
		isRavine[i] = new bool[sizeX];
		isEdge[i] = new bool[sizeX];
	}
}

void RangeImage::computeNormal()
{
	for(int i=0; i<sizeY; i++)
		for(int j=0; j<sizeX; j++){
			double len = sqrt(Ix[i][j]*Ix[i][j] + Iy[i][j]*Iy[i][j] + 1.0);
			normal[i][j][0] = (float)(Ix[i][j]/len);
			normal[i][j][1] = (float)(Iy[i][j]/len);
			normal[i][j][2] = (float)(-1.0/len);
		}
}

void RangeImage::computeNabla()
{
	for(int i=0; i<sizeY; i++)
		for(int j=0; j<sizeX; j++){
			Ix[i][j] = 0.5f*(I[i][j+1] - I[i][j-1]);
			Iy[i][j] = 0.5f*(I[i+1][j] - I[i-1][j]);

			//Ix[i][j] = (I[i-1][j+1] + 2.0f*I[i][j+1] + I[i+1][j+1]
						//- I[i-1][j-1] - 2.0f*I[i][j-1] - I[i+1][j-1])/8.0;
			//Iy[i][j] = (I[i+1][j-1] + 2.0f*I[i+1][j] + I[i+1][j+1]
						//- I[i-1][j-1] - 2.0f*I[i-1][j] - I[i-1][j+1])/8.0;
		}
}

void RangeImage::reflectImage()
{
	I[-1][-1] = I[0][0];
	for(int i=0; i<sizeX; i++)
		I[-1][i] = I[0][i];

	I[-1][sizeX] = I[0][sizeX-1];
	for(i=0; i<sizeY; i++)
		I[i][-1] = I[i][0];

	I[sizeY][-1] = I[sizeY-1][0];
	for(i=0; i<sizeX; i++)
		I[sizeY][i] = I[sizeY-1][i];

	I[sizeY][sizeX] = I[sizeY-1][sizeX-1];
	for(i=0; i<sizeY; i++)
		I[i][sizeX] = I[i][sizeX-1];
}

void RangeImage::reflectNabla()
{
	Ix[-1][-1] = Ix[0][0];
	for(int i=0; i<sizeX; i++)
		Ix[-1][i] = Ix[0][i];

	Ix[-1][sizeX] = Ix[0][sizeX-1];
	for(i=0; i<sizeY; i++)
		Ix[i][-1] = Ix[i][0];

	Ix[sizeY][-1] = Ix[sizeY-1][0];
	for(i=0; i<sizeX; i++)
		Ix[sizeY][i] = Ix[sizeY-1][i];

	Ix[sizeY][sizeX] = Ix[sizeY-1][sizeX-1];
	for(i=0; i<sizeY; i++)
		Ix[i][sizeX] = Ix[i][sizeX-1];

	Iy[-1][-1] = Iy[0][0];
	for(i=0; i<sizeX; i++)
		Iy[-1][i] = Iy[0][i];

	Iy[-1][sizeX] = Iy[0][sizeX-1];
	for(i=0; i<sizeY; i++)
		Iy[i][-1] = Iy[i][0];

	Iy[sizeY][-1] = Iy[sizeY-1][0];
	for(i=0; i<sizeX; i++)
		Iy[sizeY][i] = Iy[sizeY-1][i];

	Iy[sizeY][sizeX] = Iy[sizeY-1][sizeX-1];
	for(i=0; i<sizeY; i++)
		Iy[i][sizeX] = Iy[i][sizeX-1];
}

void RangeImage::computeNablaFromNormal()
{
	for(int i=0; i<sizeY; i++)
		for(int j=0; j<sizeX; j++){
			Ix[i][j] = -normal[i][j][0]/normal[i][j][2];
			Iy[i][j] = -normal[i][j][1]/normal[i][j][2];
		}
}

void RangeImage::computeNablaNabla()
{
	for(int i=0; i<sizeY; i++)
		for(int j=0; j<sizeX; j++){
			Ixx[i][j] = 0.5f*(Ix[i][j+1] - Ix[i][j-1]);
			Ixy[i][j] = 0.5f*(Ix[i+1][j] - Ix[i-1][j]);

			Iyx[i][j] = 0.5f*(Iy[i][j+1] - Iy[i][j-1]);
			Iyy[i][j] = 0.5f*(Iy[i+1][j] - Iy[i-1][j]);

			//Ixx[i][j] = Ix[i-1][j+1] + 2.0f*Ix[i][j+1] + Ix[i+1][j+1]
						//- Ix[i-1][j-1] - 2.0f*Ix[i][j-1] - Ix[i+1][j-1];
			//Ixy[i][j] = Ix[i+1][j-1] + 2.0f*Ix[i+1][j] + Ix[i+1][j+1]
						//- Ix[i-1][j-1] - 2.0f*Ix[i-1][j] - Ix[i-1][j+1];

			//Iyx[i][j] = Iy[i-1][j+1] + 2.0f*Iy[i][j+1] + Iy[i+1][j+1]
						//- Iy[i-1][j-1] - 2.0f*Iy[i][j-1] - Iy[i+1][j-1];
			//Iyy[i][j] = Iy[i+1][j-1] + 2.0f*Iy[i+1][j] + Iy[i+1][j+1]
						//- Iy[i-1][j-1] - 2.0f*Iy[i-1][j] - Iy[i-1][j+1];
		}
}

double RangeImage::K(int i, int j)
{
	double tmp = 1.0 + Ix[i][j]*Ix[i][j] + Iy[i][j]*Iy[i][j];
	return (Ixx[i][j]*Iyy[i][j] - Ixy[i][j]*Iyx[i][j])/(tmp*tmp);
}

double RangeImage::H(int i, int j)
{
	double tmp = 1.0 + Ix[i][j]*Ix[i][j] + Iy[i][j]*Iy[i][j];
	return ((1+Ix[i][j]*Ix[i][j])*Ixx[i][j] 
			- Ix[i][j]*Iy[i][j]*(Ixy[i][j]+Iyx[i][j]) 
			+ (1+Iy[i][j]*Iy[i][j])*Iyy[i][j])/(2.0*pow(tmp, 1.5));
}

void RangeImage::computeFaceNormal()
{
	for(int i=0; i<sizeY-1; i++){
		for(int j=0; j<sizeX-1; j++){
			float v[4][3];
			v[0][0] = j;
			v[0][1] = i;
			v[0][2] = I[i][j];

			v[1][0] = j+1;
			v[1][1] = i;
			v[1][2] = I[i][j+1];

			v[2][0] = j+1;
			v[2][1] = i+1;
			v[2][2] = I[i+1][j+1];

			v[3][0] = j;
			v[3][1] = i+1;
			v[3][2] = I[i+1][j];

			float n[3];
			n[0] = (v[1][1]-v[0][1])*(v[1][2]+v[0][2]) + (v[2][1]-v[1][1])*(v[2][2]+v[1][2])
					+ (v[3][1]-v[2][1])*(v[3][2]+v[2][2]) + (v[0][1]-v[3][1])*(v[0][2]+v[3][2]);
			n[1] = (v[1][2]-v[0][2])*(v[1][0]+v[0][0]) + (v[2][2]-v[1][2])*(v[2][0]+v[1][0])
					+ (v[3][2]-v[2][2])*(v[3][0]+v[2][0]) + (v[0][2]-v[3][2])*(v[0][0]+v[3][0]);
			n[2] = (v[1][0]-v[0][0])*(v[1][1]+v[0][1]) + (v[2][0]-v[1][0])*(v[2][1]+v[1][1])
					+ (v[3][0]-v[2][0])*(v[3][1]+v[2][1]) + (v[0][0]-v[3][0])*(v[0][1]+v[3][1]);
			float len = sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2]);
			normal_f[i][j][0] = n[0]/len;
			normal_f[i][j][1] = n[1]/len;
			normal_f[i][j][2] = n[2]/len;
		}
	}
}

void RangeImage::detectFeatureAngle(float T)
{
	for(int i=0; i<sizeY; i++){
		for(int j=0; j<sizeX; j++){
			isRidge[i][j] = false;
			isRavine[i][j] = false;
		}
	}

	for(i=0; i<sizeY-2; i++){
		for(int j=0; j<sizeX-2; j++){
			float inner = normal_f[i][j][0]*normal_f[i][j+1][0] +
							normal_f[i][j][1]*normal_f[i][j+1][1] +
							normal_f[i][j][2]*normal_f[i][j+1][2];
			if(inner < T){
				float v[3];
				v[0] = 1.0f;
				v[1] = 0;
				v[2] = I[i][j+1] - I[i][j];
				float len = sqrt(v[0]*v[0] + v[2]*v[2]);

				float dot1 = (v[0]*normal_f[i][j][0] + v[2]*normal_f[i][j][2])/len;
				if(dot1 > 1.0)
					dot1 = 1;
				else if(dot1 < -1.0)
					dot1 = -1;

				float dot2 = (- v[0]*normal_f[i][j+1][0] - v[2]*normal_f[i][j+1][2])/len;
				if(dot2 > 1.0)
					dot2 = 1;
				else if(dot2 < -1.0)
					dot2 = -1;

				if(acos(dot1) + acos(dot2) < PI){
					isRidge[i][j] = true;
					isRidge[i][j+1] = true;
				}
				else{
					isRavine[i][j] = true;
					isRavine[i][j+1] = true;
				}
			}
			inner = normal_f[i][j][0]*normal_f[i+1][j][0] +
							normal_f[i][j][1]*normal_f[i+1][j][1] +
							normal_f[i][j][2]*normal_f[i+1][j][2];
			if(inner < T){
				float v[3];
				v[0] = 0;
				v[1] = 1.0f;
				v[2] = I[i+1][j] - I[i][j];
				float len = sqrt(v[1]*v[1] + v[2]*v[2]);

				float dot1 = (v[1]*normal_f[i][j][1] + v[2]*normal_f[i][j][2])/len;
				if(dot1 > 1.0)
					dot1 = 1;
				else if(dot1 < -1.0)
					dot1 = -1;

				float dot2 = (- v[1]*normal_f[i+1][j][1] - v[2]*normal_f[i+1][j][2])/len;
				if(dot2 > 1.0)
					dot2 = 1;
				else if(dot2 < -1.0)
					dot2 = -1;

				if(acos(dot1) + acos(dot2) < PI){
					isRidge[i][j] = true;
					isRidge[i+1][j] = true;
				}
				else{
					isRavine[i][j] = true;
					isRavine[i+1][j] = true;
				}
			}
		}
	}
}

void RangeImage::detectEdge(float T)
{
	for(int i=0; i<sizeY; i++){
		for(int j=0; j<sizeX; j++){
			isEdge[i][j] = false;
		}
	}

	for(i=0; i<sizeY; i++){
		for(int j=0; j<sizeX; j++){
			if(sqrt(Ix[i][j]*Ix[i][j] + Iy[i][j]*Iy[i][j]) > T)
				isEdge[i][j] = true;
			else
				isEdge[i][j] = false;
		}
	}
}

void RangeImage::copyImage(RangeImage *image)
{
	this->setSize(image->sizeX, image->sizeY);
	for(int i=0; i<sizeY; i++)
		for(int j=0; j<sizeY; j++)
			I[i][j] = image->I[i][j];
}
