// Smoother.cpp: Smoother NX̃Cve[V
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "RangeImageViewer.h"
#include "Smoother.h"

#include "math.h"
#include "RangeImage.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// \z/
//////////////////////////////////////////////////////////////////////

Smoother::Smoother()
{
	image = NULL;
}

Smoother::~Smoother()
{

}

void Smoother::setRangeImage(RangeImage *image)
{
	this->image = image;
}

void Smoother::averageNormal(int iter, float c)
{
	int sizeX = image->sizeX;
	int sizeY = image->sizeY;

	float (**normal)[3] = image->normal;
	float (**tmp)[3] = (float (**)[3])new float*[sizeY];
	for(int m=0; m<sizeY; m++)
		tmp[m] = new float[sizeX][3];
	
	for(m=0; m<iter; m++){
		image->computeNablaNabla();
		for(int i=1; i<sizeY-1; i++){
			for(int j=1; j<sizeX-1; j++){
				double total_w = 0;
				double n[3];
				n[0] = n[1] = n[2] = 0;
				for(int k = -1; k<2; k++){
					for(int l = -1; l<2; l++){
						double H = image->H(i+k,j+l);
						double w = exp(-(double)c*(2.0*H*H-image->K(i+k,j+l)));
						total_w += w;
						n[0] += w*normal[i+k][j+l][0];
						n[1] += w*normal[i+k][j+l][1];
						n[2] += w*normal[i+k][j+l][2];
					}
				}
				if((float)total_w != 0){
					tmp[i][j][0] = (float)(n[0]/total_w);
					tmp[i][j][1] = (float)(n[1]/total_w);
					tmp[i][j][2] = (float)(n[2]/total_w);
				}
				else{
					tmp[i][j][0] = normal[i][j][0];
					tmp[i][j][1] = normal[i][j][1];
					tmp[i][j][2] = normal[i][j][2];
				}
			}
		}
		for(i=1; i<sizeY-1; i++){
			for(int j=1; j<sizeX-1; j++){
				normal[i][j][0] = tmp[i][j][0];
				normal[i][j][1] = tmp[i][j][1];
				normal[i][j][2] = tmp[i][j][2];
			}
		}
		image->computeNablaFromNormal();
		image->reflectNabla();
	}
	for(int i=1; i<sizeY-1; i++){
		for(int j=1; j<sizeX-1; j++){
			double l = sqrt(normal[i][j][0]*normal[i][j][0] 
						+ normal[i][j][1]*normal[i][j][1] 
						+ normal[i][j][2]*normal[i][j][2]);
			if((float)l != 0){
				normal[i][j][0] /= l;
				normal[i][j][1] /= l;
				normal[i][j][2] /= l;
			}
			else{
				normal[i][j][0] = 0;
				normal[i][j][1] = 0;
				normal[i][j][2] = 1.0f;
			}
		}
	}
	for(m=0; m<sizeY; m++)
		delete[] tmp[m];
	delete[] tmp;
}

void Smoother::averageFaceNormal(int iter, float c)
{
	int sizeX = image->sizeX;
	int sizeY = image->sizeY;

	float (**normal_f)[3] = image->normal_f;
	float (**I) = image->I;
	float (**tmp)[3] = (float (**)[3])new float*[sizeY-1];
	for(int m=0; m<sizeY-1; m++)
		tmp[m] = new float[sizeX-1][3];
	
	for(m=0; m<iter; m++){
		for(int i=1; i<sizeY-2; i++){
			for(int j=1; j<sizeX-2; j++){
				double nP[3];
				nP[0] = normal_f[i][j][0];
				nP[1] = normal_f[i][j][1];
				nP[2] = normal_f[i][j][2];
				double len = sqrt(nP[0]*nP[0] + nP[1]*nP[1] + nP[2]*nP[2]);
				if((float)len != 0){
					nP[0] /= len;
					nP[1] /= len;
					nP[2] /= len;
				}
				else{
					tmp[i][j][0] = normal_f[i][j][0];
					tmp[i][j][1] = normal_f[i][j][1];
					tmp[i][j][2] = normal_f[i][j][2];
					continue;
				}
				double P[3];
				P[0] = 0;
				P[1] = 0;
				P[2] = 0.25*(I[i][j] + I[i][j+1] + I[i+1][j+1] + I[i+1][j]);

				double total_w = 0;
				double n[3];
				n[0] = n[1] = n[2] = 0;
				for(int k = -1; k<2; k++){
					for(int l = -1; l<2; l++){
						double nQ[3];
						nQ[0] = normal_f[i+k][j+l][0];
						nQ[1] = normal_f[i+k][j+l][1];
						nQ[2] = normal_f[i+k][j+l][2];
						len = sqrt(nQ[0]*nQ[0] + nQ[1]*nQ[1] + nQ[2]*nQ[2]);
						if((float)len != 0){
							nQ[0] /= len;
							nQ[1] /= len;
							nQ[2] /= len;
						}
						else{
							continue;
						}
						double Q[3];
						Q[0] = k;
						Q[1] = l;
						Q[2] = 0.25*(I[i+k][j+l] + I[i+k][j+l+1] + I[i+k+1][j+l+1] + I[i+k+1][j+l]);

						double dot = nP[0]*nQ[0] + nP[1]*nQ[1] + nP[2]*nQ[2];
						if(dot > 1.0)
							dot = 1.0;
						else if(dot < -1.0)
							dot = -1.0;
						double phi = acos(dot);
						double d2 = (Q[0]-P[0])*(Q[0]-P[0]) + (Q[1]-P[1])*(Q[1]-P[1]); // + (Q[2]-P[2])*(Q[2]-P[2]);
						double w = exp(-(double)c*phi*phi/(d2+1)); // - d2);

						total_w += w;
						n[0] += w*normal_f[i+k][j+l][0];
						n[1] += w*normal_f[i+k][j+l][1];
						n[2] += w*normal_f[i+k][j+l][2];
					}
				}
				if((float)total_w != 0){
					tmp[i][j][0] = (float)(n[0]/total_w);
					tmp[i][j][1] = (float)(n[1]/total_w);
					tmp[i][j][2] = (float)(n[2]/total_w);
				}
				else{
					tmp[i][j][0] = normal_f[i][j][0];
					tmp[i][j][1] = normal_f[i][j][1];
					tmp[i][j][2] = normal_f[i][j][2];
				}
			}
		}
		for(i=1; i<sizeY-2; i++){
			for(int j=1; j<sizeX-2; j++){
				normal_f[i][j][0] = tmp[i][j][0];
				normal_f[i][j][1] = tmp[i][j][1];
				normal_f[i][j][2] = tmp[i][j][2];
			}
		}
	for(i=1; i<sizeY-2; i++){
		for(int j=1; j<sizeX-2; j++){
			double l = sqrt(normal_f[i][j][0]*normal_f[i][j][0] 
						+ normal_f[i][j][1]*normal_f[i][j][1] 
						+ normal_f[i][j][2]*normal_f[i][j][2]);
			if((float)l != 0){
				normal_f[i][j][0] /= l;
				normal_f[i][j][1] /= l;
				normal_f[i][j][2] /= l;
			}
			else{
				normal_f[i][j][0] = 0;
				normal_f[i][j][1] = 0;
				normal_f[i][j][2] = 1.0f;
			}
		}
	}
	}
	for(m=0; m<sizeY-1; m++)
		delete[] tmp[m];
	delete[] tmp;
}

void Smoother::moveVertex(int iter)
{
	int sizeX = image->sizeX;
	int sizeY = image->sizeY;

	float (**normal_f)[3] = image->normal_f;
	float (**I) = image->I;
	float (**Ix) = image->Ix;
	float (**Iy) = image->Iy;
	float **tmp = new float*[sizeY];
	for(int m=0; m<sizeY; m++)
		tmp[m] = new float[sizeX];
	
	for(m=0; m<iter; m++){
		for(int i=1; i<sizeY-1; i++){
			for(int j=1; j<sizeX-1; j++){
				float p[3];
				p[0] = 0;
				p[1] = 0;
				p[2] = I[i][j];

				double shiftZ = 0;
				int valid_N = 0;
				for(int k=-1; k<1; k++){
					for(int l=-1; l<1; l++){
						float n[3];
						n[0] = normal_f[i+k][j+l][0];
						n[1] = normal_f[i+k][j+l][1];
						n[2] = normal_f[i+k][j+l][2];
						if(fabs(n[2]) < 0.05)
							continue;

						double c[3];
						c[0] = 0.5 + (double)l;
						c[1] = 0.5 + (double)k;
						c[2] = 0.25*(I[i+k+1][j+l+1] + I[i+k+1][j+l] + I[i+k][j+l] + I[i+k][j+l+1]);
						

						double inner = n[0]*(c[0]-p[0]) + n[1]*(c[1]-p[1]) + n[2]*(c[2]-p[2]);
						shiftZ += inner/n[2];
						valid_N++;
					}
				}
				if(valid_N > 0)
					tmp[i][j] = shiftZ/(float)valid_N;
				else
					tmp[i][j] = 0;
			}
		}
		for(i=1; i<sizeY-1; i++){
			for(int j=1; j<sizeX-1; j++){
				I[i][j] += tmp[i][j];
				if(I[i][j] < 0)
					I[i][j] = 0;
				else if(I[i][j] > 255)
					I[i][j] = 255;
			}
		}
	}
	for(m=0; m<sizeY; m++)
		delete[] tmp[m];
	delete[] tmp;
}

void Smoother::smoothI(int iter, float c)
{
	for(int i=0; i<iter; i++){
		image->computeFaceNormal();
		this->averageFaceNormal(1, c);
		this->moveVertex(1);
	}
	image->computeFaceNormal();
}

void Smoother::gaussSmooth(int iter, float c)
{
	int sizeX = image->sizeX;
	int sizeY = image->sizeY;

	float (**normal_f)[3] = image->normal_f;
	float (**I) = image->I;
	float **tmp = (float **)new float[sizeY];
	for(int m=0; m<sizeY; m++)
		tmp[m] = new float[sizeX];
	
	for(m=0; m<iter; m++){
		for(int i=1; i<sizeY-1; i++){
			for(int j=1; j<sizeX-1; j++){
				double p[3];
				p[0] = 0;
				p[1] = 0;
				p[2] = I[i][j];

				double newI = 0;
				double total_w = 0;
				for(int k=-1; k<2; k++){
					for(int l=-1; l<2; l++){
						double d2 = k*k + l*l;
						double w = exp(-c*d2);
						newI += w*I[i+k][j+l];
						total_w += w;
					}
				}
				tmp[i][j] = newI/total_w;
			}
		}
		for(i=1; i<sizeY-1; i++){
			for(int j=1; j<sizeX-1; j++){
				I[i][j] = tmp[i][j];
			}
		}
	}
	for(m=0; m<sizeY; m++)
		delete[] tmp[m];
	delete[] tmp;
}

void Smoother::addNoise(float c)
{
	srand(100);
	for(int i=1; i<image->sizeY-1; i++){
		for(int j=1; j<image->sizeX-1; j++){
			image->I[i][j] += c*(2.0*((double)rand())/RAND_MAX - 1.0);
		}
	}
}


//NOT USED
void Smoother::moveVertex2(int iter)
{
	int sizeX = image->sizeX;
	int sizeY = image->sizeY;

	float (**normal_f)[3] = image->normal_f;
	float (**I) = image->I;
	float (**Ix) = image->Ix;
	float (**Iy) = image->Iy;
	float **tmp = new float*[sizeY];
	for(int m=0; m<sizeY; m++)
		tmp[m] = new float[sizeX];
	
	for(m=0; m<iter; m++){
		for(int i=1; i<sizeY-1; i++){
			for(int j=1; j<sizeX-1; j++){
				if(image->isEdge[i][j]){
					/*
					double newI = 0;
					double total_w = 0;
					for(int k=-1; k<2; k++){
						for(int l=-1; l<2; l++){
							double w = exp(-100*(Ix[i+k][j+l]*Ix[i+k][j+l] + Iy[i+k][j+l]*Iy[i+k][j+l]));
							newI += I[i+k][j+l]*w;
							total_w += w;
						}
						if((float)total_w != 0){
							tmp[i][j] = newI/total_w - I[i][j];
						}
						else{
							tmp[i][j] = 0;
						}
					}*/
					tmp[i][j] = 0;
					continue;
				}

				float p[3];
				p[0] = j;
				p[1] = i;
				p[2] = I[i][j];

				double shiftZ = 0;
				int valid_N = 0;
				int ad_v[4][2] = {{j+1,i},{j,i+1},{j-1,i},{j,i-1}};
				int ad_f[4][2] = {{j,i-1},{j,i},{j-1,i},{j-1,i-1}};

				for(int k=0; k<4; k++){
					//if(image->isEdge[ad_v[k][1]][ad_v[k][0]])
						//continue;

					double n[3];
					n[0] = 0.5*(normal_f[ad_f[k][1]][ad_f[k][0]][0] 
						+ normal_f[ad_f[(k+1)%4][1]][ad_f[(k+1)%4][0]][0]);
					n[1] = 0.5*(normal_f[ad_f[k][1]][ad_f[k][0]][1] 
						+ normal_f[ad_f[(k+1)%4][1]][ad_f[(k+1)%4][0]][1]);
					n[2] = 0.5*(normal_f[ad_f[k][1]][ad_f[k][0]][2] 
						+ normal_f[ad_f[(k+1)%4][1]][ad_f[(k+1)%4][0]][2]);
					double len = sqrt(n[0]*n[0] + n[1]*n[1] + n[2]*n[2]);
					n[0] /= len;
					n[1] /= len;
					n[2] /= len;

					double c[3];
					c[0] = 0.5*(p[0] + ad_v[k][0]);
					c[1] = 0.5*(p[1] + ad_v[k][1]);
					c[2] = 0.5*(p[2] + I[ad_v[k][1]][ad_v[k][0]]);

					double inner = n[0]*(c[0]-p[0]) + n[1]*(c[1]-p[1]) + n[2]*(c[2]-p[2]);
					shiftZ += inner/n[2];
					valid_N++;
				}
				if(valid_N > 0)
					tmp[i][j] = shiftZ/(float)valid_N;
				else
					tmp[i][j] = 0;
			}
		}
		for(i=1; i<sizeY-1; i++){
			for(int j=1; j<sizeX-1; j++){
				I[i][j] += tmp[i][j];
				if(I[i][j] < 0)
					I[i][j] = 0;
				else if(I[i][j] > 255)
					I[i][j] = 255;
			}
		}
	}
	for(m=0; m<sizeY; m++)
		delete[] tmp[m];
	delete[] tmp;
}
